#!/usr/bin/env python
# encoding: utf-8
"""
scaffold_filler.py

Created by Rod Docking on 2010-06-30.
Copyright (c) 2010 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os
import subprocess
from optparse import OptionParser
from glob import iglob
import shutil

from fasta_laundry import read_fasta, FastaRecord, reverse_complement, get_single_seq, descaffold
from sam_to_oriented_fasta import write_oriented_fasta
from assess_assembly import check_scaffold_assembly

QTRIM = 3

def submit_scaffold_assembly(key, gap, seq, counter, bam_list, options,
                            forward=True, reverse=True):
    """Construct input files and perform a single local assembly"""
    
    assembly_file_root = "{0}_{1}".format(key, counter)
    
    # Read files
    read_file_fwd = assembly_file_root + "_fwd.fa"
    read_file_rev = assembly_file_root + "_rev.fa"

    # Assemblies and targets
    assembly_file_fwd = assembly_file_root + "_assembled_fwd.fa"
    assembly_file_rev = assembly_file_root + "_assembled_rev.fa"
    front_flank_name = assembly_file_root + '_front_target.fa'
    back_flank_name = assembly_file_root + '_back_target.fa'
    
    # Blat files
    fwd_front_blat_name = assembly_file_root + '_fwd_front.psl'
    fwd_back_blat_name = assembly_file_root + '_fwd_back.psl'
    rev_front_blat_name = assembly_file_root + '_rev_front.psl'
    rev_back_blat_name = assembly_file_root + '_rev_back.psl'
    
    # For each BAM file in the input list, extract read length and insert size
    #   then extract the necessary reads
    #   - Note that we don't actually use the read length here - it's set in the options,
    #       but we DO need to use the real insert-size for each library
    #TODO - we don't need to read this file for each assembly...
    with open(bam_list, 'r') as bam_handle:
        for line in bam_handle:
            bam_file, read_length, i_size = line.split()
            read_length = int(read_length)
            i_size = int(i_size)
            i_size_fudge = i_size * 1.4

            # big_window is the range from which to extract candidate reads
            big_window = "{0}:{1}-{2}".format(
                key, max(1, int(gap[0] - i_size_fudge)), 
                min(len(seq), int(gap[1] + i_size_fudge)))

            # small_window is the range to actually reassemble
            # We extract 200bp of flanking sequence from both sides
            small_window_start = max(0, int(gap[0]-200))
            small_window_stop = min(len(seq), int(gap[1]+200))
            
            # Write the oriented fasta file - NOTE the default QTRIM here - 
            #  - need to figure out where to set this
            write_oriented_fasta(bam_file, big_window, small_window_start, 
                                small_window_stop, i_size, QTRIM, options.r_length, 
                                read_file_fwd, read_file_rev)
    
    # Bail out here if no reads were retrieved
    len_fwd = len(open(read_file_fwd).readlines())
    len_rev = len(open(read_file_rev).readlines())
    no_reads_fwd, no_reads_rev = False, False
    if forward and (len_fwd == 0):
        open(assembly_file_fwd, 'w').close()
        open(front_flank_name, 'w').close()
        no_reads_fwd = True
    if reverse and (len_rev == 0):
        open(assembly_file_rev, 'w').close()
        open(back_flank_name, 'w').close()
        no_reads_rev = True
        
    if no_reads_fwd or no_reads_rev:
        return    
    
    fwd_assembly_success, rev_assembly_success = False, False
        
    # Then assemble with DAssembler
    if forward:
        string_args = "DAssembler -r{0} -o{1} -c{2} {3}".format(
                        options.r_length, options.overlap, options.min_cov, read_file_fwd)
        args = string_args.split(" ")
        
        process = subprocess.Popen(
                         args, shell=False, cwd=os.getcwd(), stdout=subprocess.PIPE)
        
        assembly_handle = open(assembly_file_fwd, 'w')
        assembly_handle.write(">forward_assembly\n")
        assem_counter = 0
        for line in process.stdout:
            assembly_handle.write(line)
            assem_counter += 1
        assembly_handle.close()
        if assem_counter >= 1:
            fwd_assembly_success = True
    
    # Assemble the reverse direction
    if reverse:
        string_args = "DAssembler -r{0} -o{1} -c{2} {3}".format(
                        options.r_length, options.overlap, options.min_cov, read_file_rev)
        args = string_args.split(" ")
        
        process = subprocess.Popen(
                         args, shell=False, cwd=os.getcwd(), stdout=subprocess.PIPE)
        
        assembly_handle = open(assembly_file_rev, 'w')
        assembly_handle.write(">reverse_assembly\n")
        assem_counter = 0
        for line in process.stdout:
            assembly_handle.write(line)
            assem_counter += 1
        assembly_handle.close()
        
        if assem_counter >= 1:
            rev_assembly_success = True
            
        # Reverse-complement to put new sequence in assembly orientation
        assem = get_single_seq(assembly_file_rev)
        if assem:
            assembly_handle = open(assembly_file_rev, 'w')
            assembly_handle.write(">reverse_assembly\n{0}\n".format(reverse_complement(assem)))
            assembly_handle.close()            
    
    # Extract the flank sequence and write a temp file
    # Set targets as 1000bp up and downstream of gap
    if forward:
        front_flank_start = max(0, int(gap[0] - 1000))
        print "Front flank: ", front_flank_start, gap[0] 
        target_seq = seq[front_flank_start:int(gap[0])]
        target_seq = descaffold(target_seq, options.min_gap)[-1]

        front_target = open(front_flank_name, 'w')
        front_target.write(">{0}_{1}_target_front\n{2}\n".format(
            key, counter, target_seq))
        front_target.close()

    if reverse:
        
        # Exception for front-of-contig extensions:
        back_flank_stop = min(len(seq), int(gap[1] + 1000))
        print "Back flank: ", int(gap[1]), back_flank_stop
        if int(gap[1]) == -1:
            target_seq = seq[:back_flank_stop]
        else:    
            target_seq = seq[int(gap[1])+1:back_flank_stop]
        target_seq = descaffold(target_seq, options.min_gap)[0]

        back_target = open(back_flank_name, 'w')
        back_target.write(">{0}_{1}_target_back\n{2}\n".format(
            key, counter, target_seq))
        back_target.close()
    
    # Forward only - just the forward assembly against the front flank
    if forward and fwd_assembly_success:
        # BLAT front flank vs. forward assembly
        run_blat(front_flank_name, assembly_file_fwd, fwd_front_blat_name)
    
    # Reverse only - just the reverse assembly against the back flank
    if reverse and rev_assembly_success:
        # BLAT back flank vs. reverse assembly
        run_blat(back_flank_name, assembly_file_rev, rev_back_blat_name)
    
    # Gap-filling assemblies: both assemblies against both flanks
    if forward and reverse:
    
        # BLAT back flank vs. forward assembly
        if fwd_assembly_success:
            run_blat(back_flank_name, assembly_file_fwd, fwd_back_blat_name)
    
        # BLAT front flank vs. reverse assembly
        if rev_assembly_success:
            run_blat(front_flank_name, assembly_file_rev, rev_front_blat_name)
    
    return

def run_blat(database, query, output):
    """
    Run blat with named database, query, and output, and wait until finished
    """
    
    blat_process = subprocess.check_call(['blat', '-noHead', database,
                                    query, output],
                                    shell=False, cwd=os.getcwd(),
                                    stdout=open('/dev/null', 'w'))

    return

def pool_manager(args):
    """
    Manage the multiprocess pool for assemblies to submit
    """
    
    key, gap_dict, seq_dict, bam_list, options = args

    # Retrieve assembly information from dictionary
    extend_front, gap_list, extend_back = gap_dict[key]
    counter = 0
    seq = seq_dict[key]
    
    # Extract the original contig
    contig_handle = open("{0}_original.fa".format(key),'w')
    contig_handle.write(">{0}\n{1}\n".format(key, seq))
    contig_handle.close()
    
    print key, len(seq), '\n----------------'
    print gap_dict[key]
    
    # Extend front if we need to
    if extend_front:
        print "Extend front!"
        gap = extend_front
        submit_scaffold_assembly(key, gap, seq, counter, 
                                bam_list, options, forward=False)
        counter += 1
    
    # For each gap, assemble
    for gap in gap_list:
        print "Gap: ", gap
        submit_scaffold_assembly(key, gap, seq, counter, 
                                bam_list, options)
        counter += 1
        #print "\n\n\n", key, "{0}_{1}\n----------------------".format(key, counter)
    
    # Extend back if we need to
    if extend_back:
        print "Extend back!"
        gap = extend_back
        submit_scaffold_assembly(key, gap, seq, counter, 
                                bam_list, options, reverse=False)
        counter += 1

    print "Returning for {0}".format(key)
    return

def parse_options():
    """Parse options provided on the command line"""
    
    usage = "usage: %prog [options] reference.fa bam_files.txt assembly_jobs_list.txt result.txt"
    parser = OptionParser(usage)
    
    parser.add_option("-q", "--qual_trim",
        action="store", type="int", dest="qtrim",
        default=3, help="phred quality threshold for read trimming (assumes ASCII-33) [3]")
    parser.add_option("-o", "--overlap",
        action="store", type="int", dest="overlap",
        default=20, help="DAssembler overlap parameter [20]")
    parser.add_option("-c", "--min_cov",
        action="store", type="int", dest="min_cov",
        default=2, help="DAssembler minimum coverage parameter [2]")
    # parser.add_option("-i", "--i_size",
    #     action="store", type="int", dest="i_size",
    #     default=250, help="Mean fragment size [250]")
    parser.add_option("-r", "--read_length",
        action="store", type="int", dest="r_length",
        default=100, help="Read length (longer reads will be trimmed) [100]")
    parser.add_option("-n", "--min_gap",
        action="store", type='int', dest="min_gap",
        help="Minimum gap size to consider for assembly [2]",
        default=2)
        
    (options, args) = parser.parse_args()
    if not len(args) == 4:
        parser.error("incorrect number of arguments")
    
    # Replace arguments with full path names
    pathed_args = [os.path.abspath(arg) for arg in args]
    
    # Check that all files exist
    for arg in pathed_args[:2]:
        if not os.path.exists(arg):
            parser.error("{0} does not exist!".format(arg))
    
    return options, pathed_args

def main():
    """Find scaffolds and attempt to fill them!"""
    
    # Parse command-line options
    options, args = parse_options()
    reference, bam_list, assembly_jobs_list, outfile = args
        
    # Read input file to get gaps to assemble
    gap_dict = {}
    with open(assembly_jobs_list) as job_handle:
        for line in job_handle:
            # Input file is contig, list, list, list
            c = line.split('\t')
            gap_dict[c[0]] = eval(c[1]), eval(c[2]), eval(c[3].rstrip())
    
    # Retrieve the sequences of interest as a dictionary
    seq_dict = {}
    for F in read_fasta(reference):
        if F.header_id in gap_dict:
            seq_dict[F.header_id] = F.long_seq
        
    # Set up params as a list of tuples - pool.map takes a single iterable
    # Submit all the assemblies we need to do
    params = [(key, gap_dict, seq_dict, bam_list, options) for key in gap_dict]
    map(pool_manager, params)
    
    contig_counter = 0
    result_dict = {'num_gap_merge' : 0, 'num_gap_extend' : 0, 'num_gap_fail' : 0,
                   'num_end_extend' : 0, 'num_end_fail' : 0}
    results_handle = open("{0}.result".format(assembly_jobs_list),'w')
    
    # Inspect all the new assemblies 
    for key in gap_dict:
        
        print "\n\n----------------------------\nChecking assemblies for: ", key
        print gap_dict[key]
        
        # Retrieve gap information and sequences
        extend_front, gap_list, extend_back = gap_dict[key]
        seq = seq_dict[key]
        counter = 0
                
        # Split the contig into subcontigs
        subcontig_list = descaffold(seq, min_gap = options.min_gap)
        subcontig_list.reverse()

        # Start piecing together the new contig
        new_contig_seq = subcontig_list.pop()
        
        # Write the novel sequences to a file
        novel_out = open("{0}_novel.fa".format(key),'w')
                
        # Check front extension assembly if necessary
        if extend_front:
            print "\nFront-of-contig extension assembly:"
            # print "Contig length: {0} Subcontig list length: {1}".format(
            #         len(new_contig_seq), len(subcontig_list))
            gap = extend_front
            new_contig_seq, result_dict, output_string, novel_seq = check_scaffold_assembly(key,
                                                                    gap, new_contig_seq,
                                                                    subcontig_list, result_dict, counter,
                                                                    forward=False)

            results_handle.write("{0}_{1}\t{2}\t{3}\n".format(key, counter, gap, output_string))
            if novel_seq:
                novel_out.write(">{0}_{1}\n{2}\n".format(key, counter, novel_seq))
            counter += 1
        
        # Inspect the assembly, and return the growing contig sequence
        for gap in gap_list:
            print "\nChecking gap: {0} ({1}bp), ({2} of {3})".format(
                gap, gap[1]-gap[0],gap_list.index(gap) + 1, len(gap_list))
            # print "Contig length: {0} Subcontig list length: {1}".format(
            #         len(new_contig_seq), len(subcontig_list))
            new_contig_seq, result_dict, output_string, novel_seq = check_scaffold_assembly(key,
                                                                    gap, new_contig_seq,
                                                                    subcontig_list, result_dict, counter)

            results_handle.write("{0}_{1}\t{2}\t{3}\n".format(key, counter, gap, output_string))
            if novel_seq:
                novel_out.write(">{0}_{1}\n{2}\n".format(key, counter, novel_seq))
            counter += 1
        
        # Check back extension assembly if necessary
        if extend_back:
            print "\nEnd-of-contig extension assembly"
            # print "Contig length: {0} Subcontig list length: {1}".format(
            #         len(new_contig_seq), len(subcontig_list))
            gap = extend_back
            new_contig_seq, result_dict, output_string, novel_seq = check_scaffold_assembly(key,
                                                                    gap, new_contig_seq,
                                                                    subcontig_list, result_dict, counter,
                                                                    reverse=False)

            results_handle.write("{0}_{1}\t{2}\t{3}\n".format(key, counter, gap, output_string))
            if novel_seq:
                novel_out.write(">{0}_{1}\n{2}\n".format(key, counter, novel_seq))
            counter += 1
            
        # Write the new contig sequence to a file
        contig_out = open("{0}_final.fa".format(key),'w')
        contig_out.write(">{0}\n{1}\n".format(key, new_contig_seq))
        contig_out.close()
        
        # Close novel sequence handle
        novel_out.close()
        
    print "Contigs checked: ", contig_counter
    print result_dict
    
    # Print just this line to the main output file for result tracking
    out_handle = open(outfile, 'w')
    out_handle.write("ANCHOR_3:\t{0}\t{1}\t{2}\t{3}\t{4}\n".format(
        result_dict['num_gap_merge'], result_dict['num_gap_extend'], 
        result_dict['num_gap_fail'], result_dict['num_end_extend'],
        result_dict['num_end_fail']))
    out_handle.close()
    
    # Concatenate novel sequences
    novel_seq_handle = open('all_novel_seqs.fa', 'w')
    for novel_seq in iglob('*_novel.fa'):
        shutil.copyfileobj(open(novel_seq, 'r'), novel_seq_handle)
    novel_seq_handle.close()        

    # And all the sequences flanking scaffold gaps
    flank_seq_handle = open('all_flank_seqs.fa', 'w')
    for target_seq in iglob('*_target.fa'):
        shutil.copyfileobj(open(target_seq, 'r'), flank_seq_handle)
    flank_seq_handle.close()

if __name__ == '__main__':
    main()
