#!/usr/bin/env python
# encoding: utf-8
"""
assembly_parsers.py

Created by Rod Docking on 2011-07-25.
Copyright (c) 2011 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os
import re
from glob import glob

from fasta_laundry import read_fasta, FastaRecord
from alignment_runner import samtools_faidx, bwa_index, bwa_aln, bwa_samse
from sam_record import SAM_record
from cluster_job import Cluster_job, remove_submission_files

def clean_fasta_headers(output_root, current_assembly, logger):
    """
    Sanitize fasta headers - remove long header comments that break samtools faidx
    """

    # Set local variables and handles
    logger.debug("Start cleaning contig headers from {0}".format(
        current_assembly))
    
    cleaned_assembly = '{0}.clean.fa'.format(output_root)
    cleaned_handle = open(cleaned_assembly,'w')
    
    abyss_pattern = re.compile("^(\d+) (\d+)")
    
    # For each fasta entry, return just the id and sequence
    for F in read_fasta(current_assembly):
        
        # Try to capture and retain ABySS-style headers if present
        search = re.search(abyss_pattern, F.header_comment)
        if search: 
            cleaned_handle.write(">{0} {1} {2}\n{3}\n".format(
                F.header_id, search.groups()[0], search.groups()[1],
                F.long_seq))

        # Otherwise just strip to the ID
        else:
            cleaned_handle.write(">{0}\n{1}\n".format(
                F.header_id, F.long_seq))
    
    cleaned_handle.close()
    
    # Log information
    logger.debug("Finished cleaning contig headers - returning {0}".format(cleaned_assembly))

    return cleaned_assembly

def trim_short_contigs(output_root, current_assembly, threshold, logger):
    """
    Trim short contigs from the assembly
    """
    
    # Set local variables and handles
    logger.debug("Start trimming short contigs from {0} at threshold {1}".format(
        current_assembly, threshold))
    
    trimmed_assembly = '{0}.trim.fa'.format(output_root)
    trimmed_handle = open(trimmed_assembly,'w')
    trim_counter, keep_counter = 0, 0
    
    # For each fasta entry, trim if shorter than threshold
    for F in read_fasta(current_assembly):
        if len(F.long_seq) < threshold:
            trim_counter += 1
        else:
            trimmed_handle.write(F.print_fasta())
            keep_counter += 1
    
    trimmed_handle.close()
    
    # Log information
    logger.info("Trimmed {0} contigs shorter than {1}bp, and kept {2} contigs".format(
        trim_counter, threshold, keep_counter))
    logger.debug("Finished trimming short contigs - returning {0}".format(trimmed_assembly))
    
    return trimmed_assembly

def filter_short_exact_matches(output_root, current_assembly, threshold, logger):
    """
    Filter short contigs with exact matches from the assembly
    """
    
    # Set local variables and handles
    logger.debug("Filtering short exact-matching contigs in {0} at threshold {1}".format(
        current_assembly, threshold))
    
    short_contigs = '{0}.short.fa'.format(output_root)
    long_contigs = '{0}.long.fa'.format(output_root)
    short_handle = open(short_contigs,'w')
    long_handle = open(long_contigs,'w')
    short_counter, long_counter = 0, 0
    
    # Separate short from long contigs
    for F in read_fasta(current_assembly):
        if len(F.long_seq) < threshold:
            short_counter += 1
            short_handle.write(F.print_fasta())
        else:
            long_counter += 1
            long_handle.write(F.print_fasta())
    
    short_handle.close()
    long_handle.close()
    logger.info("Found {0} contigs shorter than {1}bp, and {2} longer".format(
        short_counter, threshold, long_counter))
    
    # Index long contigs
    C = Cluster_job(jobname = 'filter_{0}.sh'.format(output_root))
    C.append(samtools_faidx(long_contigs))
    C.append(bwa_index(long_contigs))
    
    # Align short contigs to long ones
    sai_out = '{0}.short.sai'.format(output_root)
    sam_file = '{0}.short.sam'.format(output_root)
    
    C.append(bwa_aln(long_contigs, short_contigs, sai_out, threads='1'))
    C.append(bwa_samse(long_contigs, sai_out, short_contigs, sam_file))
    C.run_local()
    
    # Check for perfect matches
    perfect_match_list = []
    with open(sam_file) as sam_handle:
        for line in sam_handle:
            S = SAM_record(line)
            if S.header:
                continue
            
            # Find perfect matches and add query names to list
            if S.perfect_match(edit_distance_threshold = 0):
                perfect_match_list.append(S.qname)
    
    logger.info("Found {0} exactly-matching small contigs to remove".format(len(perfect_match_list)))
    
    # Remove perfect matches from main assembly file
    filtered_assembly = '{0}.filter.fa'.format(output_root)
    filter_handle = open(filtered_assembly, 'w')
    filter_counter, keep_counter = 0, 0
    
    for F in read_fasta(current_assembly):
        if F.header_id in perfect_match_list:
            filter_counter += 1
        else:
            filter_handle.write(F.print_fasta())
            keep_counter += 1
    
    # Check that we removed the right number of contigs
    assert(filter_counter == len(perfect_match_list))
    assert(keep_counter + filter_counter == short_counter + long_counter)
    logger.info("Filtered {0} exactly-matching small contigs".format(filter_counter))
    
    # Clean up intermediate files
    remove_submission_files([C])
    for pattern in ['{0}.short.*', '{0}.long.*']:
        intermediate_files = glob(pattern.format(output_root))
        for i in intermediate_files:
            os.unlink(i)
    
    # Return new current assembly
    logger.debug("Finished filtering short exact matches, returning {0}".format(
        filtered_assembly))
    
    return filtered_assembly

def main():
    pass


if __name__ == '__main__':
    main()

