#!/usr/bin/env python
# encoding: utf-8
"""
collate_results.py

Created by Rod Docking on 2011-08-05.
Copyright (c) 2011 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os
import re

def parse_flagstat(align_stats, line, root_dir):
    """
    Parse all the flagstat files and sum stats
    """
        
    # Extract list of flagstat files and convert to list
    search = re.search("Alignment statistics in (.+)", line)

    flag_files = eval(search.groups()[0])
    flag_files = [os.path.join(root_dir, f) for f in flag_files]
    
    total_pattern = re.compile("(\d+) \+ \d+ in total")
    mapped_pattern = re.compile("(\d+) \+ \d+ mapped")
    paired_pattern = re.compile("(\d+) \+ \d+ properly paired")    
    singleton_pattern = re.compile("(\d+) \+ \d+ singletons")
    
    # Parse each flagstat file
    for flag_file in flag_files:
        for line in open(flag_file).readlines():
            
            search = None
            search = re.search(total_pattern, line)
            if search:
                num_total = int(search.groups()[0])
                continue

            search = re.search(mapped_pattern, line)
            if search:
                num_mapped = int(search.groups()[0])
                continue
            
            search = re.search(paired_pattern, line)
            if search:
                num_paired = int(search.groups()[0])
                continue

            search = re.search(singleton_pattern, line)
            if search:
                num_singleton = int(search.groups()[0])
                continue
            
        align_stats['total'] += num_total
        align_stats['mapped'] += num_mapped
        align_stats['paired'] += num_paired
        align_stats['singletons'] += num_singleton    

    return align_stats

def main():
    """
    For all the .log files on the command line, output formatted results
    """

    # Print header row
    fields = ["Trim", "Filter", "Mapped (%)", "Paired (%)", "Singletons (%)",
             "Hom. SNP", "Hom. Indel", "Gaps (bp)", "Att. Ext.",
             "Gap_merge", "Gap_ext", "Gap_fail", "Ext_pass", "Ext_fail",
             "Assembly"]
    print "||" + "||".join(fields)
    
    # For each log file, parse and print results
    for log_file in sys.argv[1:]:
    
        root_dir = os.path.dirname(log_file)
    
        # Default values
        (num_trim, num_filter, num_mapped, pct_mapped, num_paired, pct_paired, 
        num_singleton, pct_singleton, hom_snp, hom_indel, num_gaps, gap_bp,
          att_ext, gap_merge, gap_ext, gap_fail, ext_pass, ext_fail, assembly) = (
          "n/a", "n/a", 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, log_file) 
        
        # Read each log file
        with open(log_file) as log_handle:
            for line in log_handle:
                
                search = None
                
                # Trimmed contigs
                search = re.search("Trimmed (\d+) contigs shorter than", line)
                if search:
                    num_trim = search.groups()[0]
                    continue
                
                # Filtered contigs    
                search = re.search("Filtered (\d+) exactly-matching small contigs", line)
                if search:
                    num_filter = search.groups()[0]
                    continue
                
                # Parse flagstat files
                if "Alignment statistics" in line:
                    align_stats = {r:0 for r in ['total', 'mapped', 'paired', 'singletons']}
                    align_stats = parse_flagstat(align_stats, line, root_dir)
                    
                    total_reads = align_stats['total']
                    num_mapped = align_stats['mapped']
                    pct_mapped = float(num_mapped) / float(total_reads) * 100

                    num_paired = align_stats['paired']
                    pct_paired = float(num_paired) / float(total_reads) * 100
                    
                    num_singleton = align_stats['singletons']
                    pct_singleton = float(num_singleton) / float(total_reads) * 100
                    continue
                
                # SNPs and indels
                search = re.search("Correcting (\d+) homozygous positions, and (\d+) indel positions", line)
                if search:
                    hom_snp, hom_indel = search.groups()
                    continue
                
                search = re.search("(\d+) distinct gaps in \d+ distinct contigs, with (\d+) bases in gaps", line)
                if search:
                    num_gaps, gap_bp = search.groups()
                
                search = re.search("Attempting (\d+) contig extensions", line)
                if search:
                    att_ext = search.groups()[0]
                    continue

                search = re.search("Anchor results - Gap merge: (\d+) Gap extend: (\d+) Gap fail: (\d+) End extend: (\d+) End fail: (\d+)",line)
                if search:
                    gap_merge, gap_ext, gap_fail, ext_pass, ext_fail = search.groups()
                    continue
                    
        # Convert to string and print
        result = "| {0} | {1} | {2:,d} ({3:.2f})% | {4:,d} ({5:.2f})% | {6:,d} ({7:.2f})% | {8} | {9} | {10} ({11}bp) | {12} | {13} | {14} | {15} | {16} | {17} | {18}".format(
            num_trim, num_filter, num_mapped, pct_mapped, num_paired, pct_paired,
            num_singleton, pct_singleton, hom_snp, hom_indel, num_gaps, gap_bp,
            att_ext, gap_merge, gap_ext, gap_fail, ext_pass, ext_fail, assembly)
        print result
            
if __name__ == '__main__':
    main()

