#!/usr/bin/env python
# encoding: utf-8
"""
variant_runner.py

Created by Rod Docking on 2011-07-25.
Copyright (c) 2011 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os
from cluster_job import Cluster_job, submit_cluster_or_local, hold_for_job_completion, remove_submission_files
from common_utils import validate_file
from fasta_laundry import build_faidx

def call_variants_and_indels(reference, bam_list_file, output_root, logger, 
                            cluster=False, cluster_opt=None):
    """
    Run samtools mpileup to call short variants and indels
    """
    
    logger.debug("Starting variant calling for alignments in {0} to reference {1}".format(
        bam_list_file, reference))
        
    # Split the reference into a number of sub-jobs if necessary
    #  i.e., split if we're running in cluster mode on a ref > 1Mb
    region_file_list = generate_regions_files(reference, output_root, logger)
    raw_bcf = "{0}.raw.bcf".format(output_root)
    flt_vcf = "{0}.flt.vcf".format(output_root)
    
    # Submit an mpileup job for each region
    all_jobs_list = []
    partial_bcf_file_list = [r + ".part.bcf" for r in region_file_list]
    
    for region_file, partial_bcf_file in zip(region_file_list, partial_bcf_file_list):
        
        C = Cluster_job(jobname="samtools_mpileup_{0}.sh".format(
                        region_file), touch=True, 
                        host=cluster_opt.host,
                        reservation = "mem_free=3.6G,mem_token=3.6G,h_vmem=3.6G",
                        queue=cluster_opt.regular_queue)
        C.append(samtools_mpileup(reference, bam_list_file,
                                    region_file, partial_bcf_file))
        all_jobs_list = submit_cluster_or_local(all_jobs_list, C, cluster)
    
    # Wait for all mpileup jobs to finish
    hold_for_job_completion(all_jobs_list, sleep_timer=60, logger=logger)
    
    C = Cluster_job(jobname = 'bcftools_{0}.sh'.format(output_root))
    
    # Concatenate all the region-specific .bcf files
    C.append(bcftools_cat(partial_bcf_file_list, raw_bcf))
    
    # Convert to vcf and run vcfutils.pl varFilter to filter variants
    C.append(vcfutils_varfilter(raw_bcf, flt_vcf))
    C.run_local()
    
    # Validate variant calls here
    [validate_file(r, min_size=5, remove=True) for r in region_file_list]
    [validate_file(b, min_size=100, remove=True) for b in partial_bcf_file_list]
    
    # Clean submission files here
    remove_submission_files(all_jobs_list)
    remove_submission_files([C])
    
    logger.debug("Finished variant calling for alignments in {0} to reference {1} - {2}".format(
        bam_list_file, reference, flt_vcf))
        
    return flt_vcf

def bcftools_cat(partial_bcf_file_list, output_file):
    """
    Concatenate all the partial bcf files
    """
    
    args = "bcftools cat {0} > {1}".format(
        " ".join(partial_bcf_file_list), output_file)
    
    return args

def generate_regions_files(reference, output_root, logger):
    """ Generate a list of regions to split mpileup jobs """
    
    region_counter, sequence_sum = 0, 0
    region_file = output_root + ".region_{0}.txt".format(region_counter)
    region_handle = open(region_file, 'w')
    region_file_list = [region_file]
    
    # Get the size of the reference file to determine split size
    ref_size = os.path.getsize(reference)
    
    # If the size of the reference is less than 1MB, use a single job
    # Between 1MB and 100MB use 1 job/Mb
    # Otherwise, cap at 100 jobs
    if ref_size <= 1000000:
        mpileup_region_size = 100000000000
    elif ref_size <= 100000000:
        mpileup_region_size = 100000000
    else:
        mpileup_region_size = ref_size / 100
            
    logger.info("Reference size is {0} - using mpileup region size of {1}".format(
        ref_size, mpileup_region_size))
    
    # Make sure the faidx file exists
    faidx = build_faidx(reference)
    
    # Split the reference into regions
    with open(reference + '.fai', 'r') as fai_handle:
        
        for line in fai_handle:
            cols = line.split()
            c_id, c_len = cols[0], int(cols[1])
            
            # Open a new file if we need to
            if not region_handle:
                region_file = output_root + ".region_{0}.txt".format(region_counter)
                region_handle = open(region_file, 'w')
                region_file_list.append(region_file)
            
            # NOTE: this requires at least samtools 933:176 (treating mpileup -l like a BED file)
            # i.e., at least samtools 0.1.15
            region_handle.write("{0}\t{1}\t{2}\n".format(c_id, '1', c_len))
            
            # Then check if we need to start a new file
            sequence_sum += c_len
            if sequence_sum >= mpileup_region_size:
                region_handle.close()
                region_counter += 1
                sequence_sum = 0
                region_handle = None
    
    logger.info("Generated {0} region files, last is {1}.".format(
        region_counter+1, region_file))
    
    return region_file_list

def samtools_mpileup(reference, bam_list_file, region_file, partial_bcf_file):
    """ Run samtools mpileup to call variants"""
    
    return """samtools mpileup -E -d100000 -L100000 -Augf {0} -b {1} -l {2} | \\
                bcftools view -bvcg - > {3}""".format(reference,
                bam_list_file, region_file, partial_bcf_file)

def vcfutils_varfilter(input_file, output_file):
    """ Run vcfutils.pl varFilter to filter variant calls """
    
    args = "bcftools view {0} | vcfutils.pl varFilter -p -2 0 -4 0 -d 4 -w 5 > {1}".format(
        input_file, output_file)
    
    return args

def main():
	pass


if __name__ == '__main__':
	main()

