#!/usr/bin/env python
# encoding: utf-8
"""
generate_test_set.py

Created by Rod Docking on 2011-03-25.
Copyright (c) 2011 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os
import random
import gzip
import subprocess
from optparse import OptionParser

sys.path.append(os.path.join(os.path.dirname(sys.argv[0]), ".."))
from lib.fasta_laundry import reverse_complement

COVERAGE = 20
FRAGMENT_LENGTH_ONE = 300
FRAGMENT_LENGTH_TWO = 500
READ_LENGTH = 100
SCAFFOLD_LENGTH_CLOSABLE = 100           # scaffolds of N's placed WITHIN contigs - closable first round
SCAFFOLD_LENGTH_UNCLOSABLE = 500         # scaffolds of N's placed WITHIN contigs - NOT closable first round
#SCAFFOLD_LENGTH_NEGATIVE = -50          # scaffold edges have an overlap, but a scaffold is there
GAP_LENGTH = 1000                        # gaps BETWEEN scaffolded contigs

def parse_options():
    """Parse options provided on the command line"""
    
    usage = "usage: %prog [options] ref.fa OUTPUT_ROOT"
    parser = OptionParser(usage)
    
    parser.add_option("-u",
        action="store", type="float", dest="point",
        help="Per-base rate of point mutations to induce [0.001]",
        default = 0.001)
    parser.add_option("-i",
        action="store", type="float", dest="indel",
        help="Per-base rate of indels to induce [0.0001]",
        default = 0.0001)
    parser.add_option("-s",
        action="store", type="int", dest="scaffold",
        help="Number of scaffolds to introduce [20]",
        default = 20)
    parser.add_option("-c",
        action="store", type="int", dest="contigs",
        help="Number of contigs to split to [10]",
        default = 10)
    
    (opt, args) = parser.parse_args()
    
    if not (len(args) == 2):
        parser.error("incorrect number of arguments")
    
    return opt, args

def generate_fragments(library_length, output_root, contig_seq):
    """  
    Generate fake reads and output as BAM and gzipped fastq
    """

    num_fragments = (len(contig_seq) * COVERAGE) / (READ_LENGTH * 2)
    sys.stderr.write("Generating {0} fragments for {1}x coverage\n".format(num_fragments, COVERAGE))
    
    # Set up processes for writing BAM and fastq
    bam_proc = subprocess.Popen(['samtools', 'view', '-bS', '-'], 
                shell=False, cwd=os.getcwd(), 
                stdout=open('{0}_{1}.bam'.format(output_root, library_length), 'wb'),
                stdin=subprocess.PIPE)
    
    fastq_one = gzip.open('{0}_{1}_1.fq.gz'.format(output_root, library_length), 'wb')
    fastq_two = gzip.open('{0}_{1}_2.fq.gz'.format(output_root, library_length), 'wb')
        
    # Write BAM header    
    bam_proc.stdin.write("@HD\tSO unsorted\n")
    bam_proc.stdin.write("@SQ\tSN: 1\tLN: 1\n")
    
    # Generate reads
    for i in range(num_fragments):
        
        frag_loc = random.randrange(len(contig_seq) - library_length - 10)
        frag_len = int(random.normalvariate(library_length, 10))
        
        fwd_seq = contig_seq[frag_loc:frag_loc + READ_LENGTH]
        rev_seq = contig_seq[frag_loc + frag_len - READ_LENGTH : 
                            frag_loc + frag_len]
        
        if not (len(fwd_seq) == len(rev_seq) == READ_LENGTH):
            continue
        
        if i % 2 == 0:
            fwd_seq = reverse_complement(fwd_seq)
        else:
            rev_seq = reverse_complement(rev_seq)    
        
        # Write as BAM
        sam_one = "read_{0}\t68\t*\t0\t0\t*\t*\t0\t0\t{1}\t{2}\n".format(
                   i, fwd_seq, "I"*READ_LENGTH)
        sam_two = "read_{0}\t132\t*\t0\t0\t*\t*\t0\t0\t{1}\t{2}\n".format(
                   i, rev_seq, "I"*READ_LENGTH)                 
        bam_proc.stdin.write(sam_one)
        bam_proc.stdin.write(sam_two)
        
        # Write as fastq
        fastq_one.write("@read_{0}/1\n{1}\n+\n{2}\n".format(i, fwd_seq, "I"*READ_LENGTH))
        fastq_two.write("@read_{0}/2\n{1}\n+\n{2}\n".format(i, rev_seq, "I"*READ_LENGTH))
        
    fastq_one.close()
    fastq_two.close()
    
    # Write input files
    fq_input_file = open('{0}.fastq.in'.format(output_root), 'a')
    fq_input_file.write("{0}_{1}_1.fq.gz\n{0}_{1}_2.fq.gz\n".format(output_root, library_length))
    fq_input_file.close()
    
    bam_input_file = open('{0}.bam.in'.format(output_root), 'a')
    bam_input_file.write("{0}_{1}.bam\n".format(output_root, library_length))
    bam_input_file.close()
            
    return

def induce_single_N(seq):
    """ 
    Induce a point mutation at a random base in the sequence 
    """

    position = random.randrange(len(seq))
    return seq[:position] + 'N' + seq[position+1:]

def induce_point_mutation(seq, opt):
    """ 
    Induce a point mutation at a random base in the sequence 
    """

    bases = ['A','C','G','T']
    position = random.randrange(len(seq))
    current_base = new_base = seq[position]

    while current_base == new_base:
        new_base = random.choice(bases)

    # Output as pseudo-vcf - the original base should be the ALT, 
    # since that's what the reads will have
    sys.stderr.write("I\t{0}\t.\t{1}\t{2}\n".format(position, new_base, current_base))

    return seq[:position] + new_base + seq[position+1:]
    
def induce_indel(seq, opt):
    """ 
    Induce an indel at a random base in the sequence 
    """

    bases = ['A','C','G','T']
    position = random.randrange(len(seq))
    
    # Randomly choose insertion or deletion
    if random.random() >= 0.5:
        seq = seq[:position] + random.choice(bases) + seq[position:]    # insertion
        sys.stderr.write("I\t{0}\t.\t{1}\t{2}\tINDEL\t(insertion)\t{3}\n".format(
                        position, seq[position-1:position+2], seq[position-1:position+3], len(seq)))
    else:
        seq = seq[:position] + seq[position+1:]                         # deletion        
        sys.stderr.write("I\t{0}\t.\t{1}\t{2}\tINDEL\t(deletion)\t{3}\n".format(
                        position, seq[position-1:position+2], seq[position-1:position+1], len(seq)))

    return seq    

def induce_scaffold(seq):
    """ 
    Induce a scaffold at a random base in the sequence 
    """
    
    # Prevent a scaffold at the very beginning or end of the sequence
    position = random.randrange(len(seq)-2000) + 1000
    sys.stderr.write("I\t{0}\tSCAFFOLD\t{1}\n".format(position, len(seq)))
    
    # Randomly choose a 'closable', 'unclosable', or 'negative' scaffold length
    # scaff_size = random.choice(
    #     [SCAFFOLD_LENGTH_CLOSABLE, SCAFFOLD_LENGTH_UNCLOSABLE, SCAFFOLD_LENGTH_NEGATIVE])
    scaff_size = random.choice(
        [SCAFFOLD_LENGTH_CLOSABLE, SCAFFOLD_LENGTH_UNCLOSABLE])
    #scaff_size = SCAFFOLD_LENGTH_CLOSABLE
    
    if scaff_size > 0:
        res = seq[:position] + 'N'*scaff_size + seq[position+scaff_size:]
    else:    
        res = seq[:position] + 'N'*-1*scaff_size + seq[position+scaff_size:]
    
    return res

def generate_short_contigs(n, contigs):
    """
    Generate short, perfectly-matching contigs
    """
    
    short_contigs = []
    for i in range(n):
        contig = random.choice(contigs)
        start = random.randint(0, len(contig)-100)
        print contig[start:start+50]
        short_contigs.append(contig[start:start+50])
    
    return short_contigs

def random_header():
    """
    Return a long string to mimic ABySS-style headers
    """

    id_list = [random.randrange(1000) for a in range(1000)]
    sign_list = [random.choice(['+','-']) for a in range(1000)]

    header_list = ["{0}{1}".format(a,b) for a,b in zip(id_list, sign_list)]

    return ",".join(header_list)

def main():
    """ 
    From a reference sequence, introduce point mutations, indels, and scaffolds 
    """

    # Parse options
    opt, args = parse_options()
    input_file, output_root = args

    # Read the original sequence - this will concatenate any multi-entry FASTA files
    contig_seq_list = []
    with open(input_file, 'r') as fh:
        for line in fh:
            if ">" in line:
                sys.stdout.write(line)
            else:    
                contig_seq_list.append(line.rstrip())
    
    contig_seq = "".join(contig_seq_list).upper()
    sys.stderr.write("Contig is {0}bp\n".format(len(contig_seq)))            
    
    # Sample reads from the unmodified contigs
    for library_length in [FRAGMENT_LENGTH_ONE, FRAGMENT_LENGTH_TWO]:
        generate_fragments(library_length, output_root, contig_seq)
    
    # Calculate the number of point mutations and indels to introduce
    num_point_mutations = int(opt.point * len(contig_seq))
    num_indels = int(opt.indel * len(contig_seq))
    sys.stderr.write("Inducing {0} point mutations and {1} indels\n".format(
                        num_point_mutations, num_indels))
    
    # Introduce random point mutations into the assembly
    for i in range(num_point_mutations):
        contig_seq = induce_point_mutation(contig_seq, opt)
    
    # Introduce random single-N characters
    for i in range(num_point_mutations):
        contig_seq = induce_single_N(contig_seq)
    
    # Introduce random indels into the assembly    
    for i in range(num_indels):
        contig_seq = induce_indel(contig_seq, opt)

    # Induce scaffolds into the assembly
    for i in range(opt.scaffold):
        contig_seq = induce_scaffold(contig_seq)
    
    # Generate break-points for new contigs
    good_set = False
    contigs = []
    counter = 0
    
    while good_set is False:
        breakpoints = [0] + [random.randint(0, len(contig_seq)) for i in range(opt.contigs-1)] + [len(contig_seq)]
        breakpoints.sort()
        contigs = []
        counter += 1
    
        for i in range(opt.contigs):
            print i, breakpoints[i], breakpoints[i+1], '==>',
        
            if not breakpoints[i] in [0, len(contig_seq)]:
                print i, breakpoints[i] + GAP_LENGTH/2, breakpoints[i+1] - GAP_LENGTH/2
                contig = contig_seq[(breakpoints[i] + GAP_LENGTH/2) : (breakpoints[i+1] - GAP_LENGTH/2)]
            else:
                print i, breakpoints[i], breakpoints[i+1]    
                contig = contig_seq[breakpoints[i]:breakpoints[i+1]]
            
            # Slice the contig and trim and leading or trailing N characters
            contig = contig.strip("N")
            contigs.append(contig)

        # Check that all contigs are above a minimum threshold size
        min_contig = min([len(c) for c in contigs])
        print "Smallest contig: ", min_contig
        
        # Check that there are no scaffolds too close to the end of a contig
        no_near_scaffolds = True
        for c in contigs:
            if ("N" in c[:100]) or ("N" in c[-100:]):
                 no_near_scaffolds = False
        
        # End if we've found a good set
        if min_contig > 1000 and no_near_scaffolds:
            good_set = True
        
        if counter >= 1000:
            print "Failed to generate a good contig set"
            sys.exit()    

    # Introduce some small, exactly-matching contigs to be subsumed
    good_set_short = False
    short_contigs = []
    while good_set_short is False:
        short_contigs = generate_short_contigs(10, contigs)
        good_set_short = True
        for s in short_contigs:
            if "N" in s:
                good_set_short = False

    print short_contigs
    contigs.extend(short_contigs)

    # Write out final contigs
    contig_handle = open('{0}-contigs.fa'.format(output_root), 'w')
    for i in range(len(contigs)):
        contig = contigs[i]
        
        # Make a long random header to break bwa
        header = random_header()
        contig_handle.write(">{0} {1} {2} {3}\n{4}\n".format(
            i,len(contig), len(contig), header, contig))
        
    

if __name__ == '__main__':
    main()

