#!/usr/bin/env python
# encoding: utf-8
"""
parse_known_duplication_in_abyss_assembly.py

Created by Rod Docking on 2010-10-22.
Copyright (c) 2010 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import operator
import string

MIN_LENGTH_TO_CHECK = 500
MAX_MISMATCH = 1000

def reverse_contig_path(path):
    """From an ABySS-style contig path list, return the reversed orientation"""
    
    flip_signs = [p.translate(string.maketrans("+-", "-+")) for p in path]
    return flip_signs[::-1]

def reverse_contig_path_string(path_string, delimiter=","):
    
    path_list = path_string.split(delimiter)
    reverse_list = reverse_contig_path(path_list)
    return ",".join(reverse_list)

def reverse_node_ori(node):
    """Just flip +/- in a node name"""
    return node.translate(string.maketrans("+-", "-+"))

def trim_focal_contig(seed, first_path):
    """Trim a single contig from a path - note that this is a simpler version of path_compare from below"""
    
    # print "\nTrimming focal contig:"
    # print "Trim {0} from {1}".format(seed, first_path)
    
    # Split to lists
    first_list = first_path.split(',')
    plus_ori_seed, minus_ori_seed = "{0}+".format(seed), "{0}-".format(seed)
    
    if plus_ori_seed in first_list:
        signed_seed = plus_ori_seed
    elif minus_ori_seed in first_list:
        signed_seed = minus_ori_seed
    else:
        print "Seed is not in list!"
        sys.exit()
    
    # Find the seed contig in both paths
    first_index = first_list.index(signed_seed)
    #print "Seed in first path: ", first_index
    
    # Return all the paths that have at least one contig:
    new_paths_raw = [first_list[:first_index], first_list[first_index+1:]]
    new_paths = [new_paths_raw[x] for x in range(len(new_paths_raw)) if (len(new_paths_raw[x]) > 0)]
    
    return new_paths

def path_compare(seed, first_path, second_path, se_length_dict):
    """Compare and find overlaps between two paths"""
    
    print "\nComparing paths:"
    print "Seed: ", seed, se_length_dict[seed]
    print "First path: ", first_path
    print "Second path: ", second_path
    
    # Split to lists
    first_list = first_path.split(',')
    second_list = second_path.split(',')
    
    plus_ori_seed, minus_ori_seed = "{0}+".format(seed), "{0}-".format(seed)
    
    # Check that the seed contig is in the same orientation in both paths
    if (plus_ori_seed in first_list) and (plus_ori_seed in second_list):
        signed_seed = plus_ori_seed
    elif (minus_ori_seed in first_list) and (minus_ori_seed in second_list):
        signed_seed = minus_ori_seed
    elif (plus_ori_seed in first_list):
        second_list = reverse_contig_path(second_list)
        signed_seed = plus_ori_seed
    elif (minus_ori_seed in first_list):
        second_list = reverse_contig_path(second_list)
        signed_seed = minus_ori_seed
    
    # Find the seed contig in both paths
    first_index = first_list.index(signed_seed)
    second_index = second_list.index(signed_seed)
    
    # print "Seed in first path: ", first_index
    # print "Seed in second path: ", second_index
    
    # From the starting index position, search forwards from the index
    #   in the first list to the end of the first path
    match_length_fwd, mismatch_length_fwd = 0, 0
    match_length_rev, mismatch_length_rev = 0, 0
    mismatch_list = []
    
    for i, j in zip(
        range(first_index, len(first_list)), range(second_index, len(second_list))):
        
        if "N" in first_list[i]:
            first_length = int(first_list[i][:-1])
        else:
            first_length = se_length_dict[int(first_list[i].rstrip('+-'))]
        
        if "N" in second_list[j]:
            first_length = int(second_list[j][:-1])
        else:
            second_length = se_length_dict[int(second_list[j].rstrip('+-'))]
        
        if first_list[i] == second_list[j]:
            first_stop, second_stop = i, j
            match_length_fwd += first_length
        else:
            print "Mismatch!", first_list[i], second_list[j]
            mismatch_length_fwd += int((first_length + second_length) / 2)
            mismatch_list.append(second_list[j])
        
        if mismatch_length_fwd > MAX_MISMATCH:
            mismatch_list = []
            break
    
    # print "Forward match length: ", match_length_fwd
    # print "Forward mismatch length: ", mismatch_length_fwd
    
    # And backwards:
    for i, j in zip(
        range(first_index, -1, -1), range(second_index, -1, -1)):
        if "N" in first_list[i]:
            first_length = int(first_list[i][:-1])
        else:
            first_length = se_length_dict[int(first_list[i].rstrip('+-'))]
        
        if "N" in second_list[j]:
            first_length = int(second_list[j][:-1])
        else:
            second_length = se_length_dict[int(second_list[j].rstrip('+-'))]
        
        #print i, first_list[i], j, second_list[j]
        if first_list[i] == second_list[j]:
            first_start, second_start = i, j
            match_length_rev += first_length
        else:
            print "Mismatch!", first_list[i], second_list[j]
            mismatch_length_rev += int((first_length + second_length) / 2)
            mismatch_list.append(second_list[j])
        
        if mismatch_length_rev > MAX_MISMATCH:
            mismatch_list = []
            break
    
    first_stop += 1
    second_stop += 1
    
    # print "Reverse match length: ", match_length_rev
    # print "Reverse mismatch length: ", mismatch_length_rev
    #
    print "\nMatch in first path:", first_start, first_stop, first_list[first_start:first_stop]
    print "Match in second path:", second_start, second_stop, second_list[second_start:second_stop]
    print "\nNew paths:"
    print "First pre-match: ", first_list[:first_start]
    print "First post-match: ", first_list[first_stop:]
    print "Match: ", first_list[first_start:first_stop]
    print "Second pre-match: ", second_list[:second_start]
    print "Second post-match: ", second_list[second_stop:]
    
    # Return all the paths that have at least one contig:
    new_paths_raw = [first_list[first_start:first_stop], first_list[:first_start], first_list[first_stop:], second_list[:second_start], second_list[second_stop:]]
    new_paths = [new_paths_raw[x] for x in range(len(new_paths_raw)) if (len(new_paths_raw[x]) > 0)]
        
    print "Returning these new paths:"
    for n in new_paths:
        print n
    
    filtered_mismatch_list = []
    for mismatch_contig in mismatch_list:
        for path in new_paths:
            if mismatch_contig in path or reverse_node_ori(mismatch_contig) in path:
                filtered_mismatch_list.append(mismatch_contig)
    
    # remove duplicate entries
    filtered_mismatch_list = list(set(filtered_mismatch_list))
    
    print "Original mismatch list: ", mismatch_list
    print "Contigs appearing in paths: ", filtered_mismatch_list
    for filtered_contig in filtered_mismatch_list:
        mismatch_list.remove(filtered_contig)
    
    print "Mismatch contigs to remove:"
    print mismatch_list
    
    return new_paths, mismatch_list

def multi_aligner(seed, signed_paths):
    """Attempt a multiple alignment of contig paths"""
    
    print "Multi-alignment!"
    print seed
    
    path_list = [p.split(',') for p in signed_paths]
    
    # Find the latest position in a path for the seed contig
    max_index = False
    offset_list = []
    for path in path_list:
        seed_index = path.index(seed)
        offset_list.append(seed_index)
        if not max_index:
            max_index = seed_index
        
        if seed_index > max_index:
            max_index = seed_index
    
    print "Max seed index", max_index
    
    # Print the alignment and calculate the 'tail' left over for each path
    max_tail = False
    tail_list = []
    print "Aligned paths:"
    for path, index in zip(path_list, offset_list):
        if index < max_index:
            print "\t"*(max_index-index),
        tail = len(path) - index
        tail_list.append(tail)
        print "\t".join(path), index, tail
        
        if not max_tail:
            max_tail = tail
        
        if tail > max_tail:
            max_tail = tail
    
    # Pad paths to put them in alignment position
    #print "Padded paths:"
    padded_path_list = []
    for path, index, tail in zip(path_list, offset_list, tail_list):
        
        padded_path = path
        for i in range(max_index - index):
            padded_path.insert(0, None)
        
        for i in range(max_tail - tail):
            padded_path.append(None)
        
        #print padded_path
        padded_path_list.append(padded_path)
    
    # Go through the aligned paths to find the longest segment where all agree
    print "Iterating through aligned contigs:"
    
    # Set to an impossible value - having trouble with 0 == False
    align_start, align_stop = -99, -99
    
    # From the seed contig, continue to the right...
    for i in range(max_index,len(padded_path_list[0])):
        # Get the contig at this position
        aligned_contigs = [padded_path_list[j][i] for j in range(len(padded_path_list))]
        
        # Collapse duplicates
        aligned_list = list(set(aligned_contigs))
        print i, aligned_list, len(aligned_list)
        
        if len(aligned_list) > 1:
            align_stop = i
            print "Align stop at ", i
            break
    
    if align_stop == -99:
        align_stop = len(padded_path_list[0])
    
    # From the seed contig, continue to the right...
    for i in range(0,max_index):
        # Get the contig at this position
        aligned_contigs = [padded_path_list[j][i] for j in range(len(padded_path_list))]
        
        # Collapse duplicates
        aligned_list = list(set(aligned_contigs))
        print i, aligned_list, len(aligned_list)
        
        if len(aligned_list) > 1:
            align_start = i+1
            print "Align start at ", i+1
        
    print align_start, align_stop, padded_path_list[0][align_start:align_stop]
    
    # Extract the paths that aren't part of the aligned portion
    front_paths = [p[0:align_start] for p in padded_path_list]
    back_paths = [p[align_stop:] for p in padded_path_list]
    
    print "Front paths: ", front_paths
    print "Back paths: ", back_paths
    
    # Add all the paths to a single list
    new_paths_raw = [padded_path_list[0][align_start:align_stop]]
    new_paths_raw.extend(front_paths)
    new_paths_raw.extend(back_paths)
    
    # Trim any remaining 'None' entries
    new_paths_trim_none = []
    for path in new_paths_raw:
        trimmed_path = []
        for p in path:
            if not p == None:
                trimmed_path.append(p)
        new_paths_trim_none.append(trimmed_path)
    
    # Trim any empty paths
    new_paths = [new_paths_trim_none[x] for x in range(len(new_paths_trim_none)) if (len(new_paths_trim_none[x]) > 0)]
    
    # Remove any duplicates
    new_paths_no_dupes = []
    [new_paths_no_dupes.append(i) for i in new_paths if not new_paths_no_dupes.count(i)]
    
    found_dup = False
    for path in new_paths_no_dupes:
        print path, new_paths_no_dupes.count(path)
        if not new_paths_no_dupes.count(path) == 1:
            print "Duplicate path!!!"
            found_dup = True
    
    if found_dup:
        sys.exit()
    
    return new_paths_no_dupes


def path_compare_multi(key, orig_paths, se_count_dict, se_length_dict):
    """Compare paths when there are more than two paths using the key single-end contig"""
    
    print "Seed contig: ", key
    
    print "Original paths:"
    signed_paths = []
    for path in orig_paths:
        print path
        path_list = path.split(',')
        print path_list
        if "{0}+".format(key) in path_list:
            signed_paths.append(path)
        else:
            signed_paths.append(",".join(reverse_contig_path(path_list)))
    
    print "Signed paths:"
    for path in signed_paths:
        print path
    
    # Call the multi-aligner
    new_paths = multi_aligner("{0}+".format(key), signed_paths)
    
    print "Returning these paths:"
    for n in new_paths:
        print n
    
    return new_paths

def find_paths_to_collapse(sorted_lengths, se_count_dict, se_length_dict, 
    se_path_dict, se_exclude_dict, pe_lookup_dict):
    """ Find new paths to write, and mark paired-end and singleton contigs for exclusion """
    
    # Contigs to exclude, new paths to write
    pe_exclude_dict = {}
    new_contig_paths = []
    
    # Go through the length-sorted list of replicated single-end contigs
    for key, value in sorted_lengths:
        
        print "START", key, value, se_count_dict[key], "x"
        
        # Check for cases to continue or break:
        
        # Break if we've passed the minimum length threshold
        if se_length_dict[key] < MIN_LENGTH_TO_CHECK:
            break
        
        # Continue if this se-contig has already been involved in a trimming operation
        already_checked = False
        for se_path in se_path_dict[key]:
            if pe_exclude_dict.has_key(pe_lookup_dict[se_path]):
                already_checked = True
        if already_checked:
            print " ==> This contig has already been involved in a trimming operation"
            continue
        
        # Continue if this contig is present less than twice in the assembly
        if se_count_dict[key] < 2:
            continue
        
        # If we make it this far, check for overlap between contig paths:
        
        # Find the longest overlapping path between the two paths
        if se_count_dict[key] == 2:
            print "\n--------------\nPairwise alignment: ", se_path_dict[key][0], se_path_dict[key][1]
            new_paths, mismatch_list = path_compare(key, se_path_dict[key][0], se_path_dict[key][1], se_length_dict)
            
            # Mismatch contigs are NOT written to the final assembly
            #   and so ARE included in the single-end exclusion dict
            for mismatch_contig in mismatch_list:
                if not "N" in mismatch_contig:
                    c_id = int(mismatch_contig.rstrip('+-'))
                    se_exclude_dict[c_id] = 1
        
        # For the multi-overlap case, try to find the longest paths available...
        elif not se_count_dict[key] == 2:
            print "\n--------------\nMulti-alignment: ", se_path_dict[key]
            new_paths = path_compare_multi(key, se_path_dict[key], se_count_dict, se_length_dict)
        
        # Mark the old pe-contigs for exclusion - this is what 'already_checked' above is based on
        for se_path in se_path_dict[key]:
            pe_exclude_dict[int(pe_lookup_dict[se_path])] = 1
            print "Excluding PE contig", int(pe_lookup_dict[se_path])
        
        # Add the new paths to write
        new_contig_paths.extend(new_paths)
    
    return pe_exclude_dict, new_contig_paths, se_exclude_dict
        
def parse_se_assembly(assembly_root):
    """
    Parse the single-end assembly fasta file
    
    In this function, we:
        - Gather the names and lengths of all single-end contigs
        - Initialize the count and path dictionaries for all single-end contigs
    """
    
    se_count_dict = {}      # Counts how many times each se-contig occurs in the pe-assembly
    se_length_dict = {}     # Contig length store
    se_path_dict = {}       # For each se-contig, store all the paired-end paths it occurs in
    
    # For each intermediate stage of the assembly
    for se_file in [assembly_root + '-3.fa', assembly_root + '-4.fa', assembly_root + '-5.fa']:
        
        # Open the assembly file and count all the contigs
        for line in open(se_file,'r'):
            
            if not ">" in line:
                continue
            
            c = line.split()
            c_id = int(c[0].lstrip('>'))
            length = int(c[1])
            se_count_dict[c_id] = 0
            se_length_dict[c_id] = length
            se_path_dict[c_id] = []
        
        print "Parsed SE contig file {0}, collected {1} contigs in total".format(
            se_file, len(se_count_dict.keys()))
    
    # Store the last contig id
    max_contig_id = c_id
    
    return se_count_dict, se_length_dict, se_path_dict, max_contig_id

def parse_pe_assembly(assembly_root, se_count_dict, se_path_dict, pe_path_file):
    """Parse the paired-end assembly fasta and paths file"""
    
    pe_lookup_dict = {}
    pe_path_dict = {}
    se_exclude_dict = {}
    
    for line in open(pe_path_file,'r'):
        
        c = line.split()
        c_id = int(c[0])
        
        # If there's only a single contig listed, it needs to be recorded
        #   so it can be re-written and excluded from the final assembly
        if len(c) == 1:
            se_exclude_dict[c_id] = 1
            continue
        
        # Find all the single-end contigs (joining with "," since the code below expects it -
        #   this was the old format for contig paths)
        se_path = ",".join(c[1:])
        se_contigs = []
        for contig in se_path.split(","):
            if not "N" in contig:
                se_contigs.append(int(contig.rstrip('+-')))
        
        # Store the single-end path as a key in the lookup dictionary
        #  pe_lookup_dict has keys of single-end paths, and values of (contig_id, length)
        pe_lookup_dict[se_path] = c_id
        pe_path_dict[c_id] = se_path
        
        # Increment counters for all the single-end contigs in this paired-end contig
        # Add this path to the dictionary for each involved single-end contig as well
        for contig in se_contigs:
            if se_count_dict.has_key(contig):
                
                # If a se-contig is included twice in the same path, only add it once
                if not se_path in se_path_dict[contig]:
                    se_count_dict[contig] += 1
                    se_path_dict[contig].append(se_path)
    
    # Store the last contig id
    max_contig_id = c_id
    
    print "Parsed PE path file {0}, found {1} SE contigs to exclude from PE assembly".format(
        pe_path_file, len(se_exclude_dict))
    print "Parsed PE path file, found {0} contigs made up of multiple SE-contigs".format(len(pe_path_dict.keys()))
    
    return se_count_dict, se_path_dict, se_exclude_dict, pe_lookup_dict, pe_path_dict, max_contig_id

def write_path_file(se_exclude_dict, pe_exclude_dict, new_contig_paths, pe_path_dict, max_contig_id, output_path_file):
    """ Write a new path file """
    
    print "SE contigs to exclude: ", len(se_exclude_dict)
    print "PE contigs to exclude: ", len(pe_exclude_dict)
    print "New contig paths:", len(new_contig_paths)
    
    # Trim leading or trailing scaffold gaps from contig paths
    for i in range(len(new_contig_paths)):

        if "N" in new_contig_paths[i][0]:
            new_contig_paths[i] = new_contig_paths[i][1:]
        
        if "N" in new_contig_paths[i][-1]:
            new_contig_paths[i] = new_contig_paths[i][:-1]

    # If a path only has a single contig in it, it doesn't need to be written to the paths file
    new_contig_paths = [path for path in new_contig_paths if len(path) > 1]

    # Find all the unmodified paired-end contigs
    pe_keeper_contigs = []
    for pe_contig in pe_path_dict:
        if pe_exclude_dict.has_key(pe_contig):
            continue
        else:
            pe_keeper_contigs.append(pe_contig)
    pe_keeper_contigs.sort()
    
    print "Remaining paired-end contigs", len(pe_keeper_contigs)
    print "New paired-end contigs", len(new_contig_paths)

    # Sort the list of single-end contigs to exclude
    print "Single-end contigs to EXCLUDE: ", len(se_exclude_dict)
    final_se_contigs_list = sorted(se_exclude_dict.keys())
    
    # Open the new paths file
    path_handle = open(output_path_file, 'w')
    
    # Write the single-end contig IDs to be excluded from the PE assembly:
    for se_contig in final_se_contigs_list:
        path_handle.write("{0}\n".format(se_contig))
    
    # Write the unedited paired-end contigs with their original ID's
    for pe_contig in pe_keeper_contigs:
        pe_path = pe_path_dict[pe_contig].split(",")
        path_string = " ".join(pe_path)
        path_handle.write("{0}\t{1}\n".format(pe_contig, path_string))

    # Set the next contig ID so we write all new contig IDs 
    next_contig_id = max_contig_id + 1
    
    # Write the new PE paths
    print "New PE contigs:"
    for pe_path in new_contig_paths:
        
        next_contig_id += 1
        path_string = (" ").join(pe_path)
        path_handle.write("{0}\t{1}\n".format(next_contig_id, path_string))
    
    path_handle.close()
    
def main():
    """From the single-end assembly file and the paired-end assembly file, count the amount
        of known duplication - single-end contigs used multiple times in the PE assembly"""
    
    assembly_root = sys.argv[1]
    pe_path_file = sys.argv[2]
    output_path_file = sys.argv[3]
    
    # Parse single-end assembly files
    se_count_dict, se_length_dict, se_path_dict, max_contig_id_se = parse_se_assembly(assembly_root)
    
    # Parse paired-end assembly file
    se_count_dict, se_path_dict, se_exclude_dict, pe_lookup_dict, pe_path_dict, max_contig_id_pe = \
        parse_pe_assembly(assembly_root, se_count_dict, se_path_dict, pe_path_file)
    
    max_contig_id = max([max_contig_id_se, max_contig_id_pe])
    print "Max contig ID: ", max_contig_id
    
    # Sort header dict by contig length from high to low
    sorted_lengths = sorted(se_length_dict.iteritems(), key=operator.itemgetter(1), reverse=True)
    
    # Print out a quick report of duplicated contigs longer than the threshold
    print "Potential contigs to collapse:"
    print "ID\tLength\tCount"
    sum_duplicated = 0
    for s in sorted_lengths:
        if s[1] > MIN_LENGTH_TO_CHECK and se_count_dict[s[0]] > 1:
            sys.stderr.write("{0}\t{1}\t{2}\n".format(
                s[0], s[1], se_count_dict[s[0]]))
            sum_duplicated += s[1] * (se_count_dict[s[0]] - 1)
    
    print "Estimated duplication: ", sum_duplicated        
    
    # Find new paths to write, and mark paired-end and singleton contigs for exclusion
    pe_exclude_dict, new_contig_paths, se_exclude_dict = \
        find_paths_to_collapse(sorted_lengths, se_count_dict, se_length_dict, 
        se_path_dict, se_exclude_dict, pe_lookup_dict)
    
    # Using the newly-generated contig paths, write a new paths file, and a new assembly file
    write_path_file(se_exclude_dict, pe_exclude_dict, new_contig_paths, pe_path_dict, max_contig_id, output_path_file)


if __name__ == '__main__':
    main()

